# Copyright 2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""Tools related to logging."""

import logging
import sys

import foffinf

logger = logging.getLogger(__name__)

_TRACE_LEVEL = 5
LOG_LEVELS = {
    "quiet": (logging.WARNING, "Set logging to be quiet."),
    "verbose": (logging.DEBUG, "Set logging to verbose."),
    "trace": (_TRACE_LEVEL, "Set logging for tracing (beware, there may be too many messages)."),
    None: (logging.INFO, None),  # default no-option-set
}


def logsetup(selected_level):
    """Prepare all logging."""
    # add trace level to logging
    logging.addLevelName(_TRACE_LEVEL, "TRACE")
    logging.TRACE = _TRACE_LEVEL

    # config
    logging.basicConfig(
        format='%(asctime)s.%(msecs)03d %(levelname)-5s %(message)s',
        datefmt='%H:%M:%S', stream=sys.stdout)

    # set proper level to the whole app
    level, _ = LOG_LEVELS[selected_level]
    logging.getLogger("nysor").setLevel(level)

    # use new format styles
    foffinf.formatize("nysor", scatter=True)


def log_notdone(msg, **items):
    """Nicely log a not implemented error."""
    reprs = [f"{k}={{!r}}" for k in items.keys()]
    fullmsg = f"Not Implemented! {msg}: {' '.join(reprs)}"
    logger.error(fullmsg, *items.values())



# Copyright 2025-2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""Receive, process, and manage all notifications from Neovim."""

import logging
from collections import defaultdict
from typing import Any

from nysor.utils import call_async

logger = logging.getLogger(__name__)


class DynamicCache:
    """A cache that is cleaned up when any of the section labels change."""

    def __init__(self):
        self._data = {}
        self._labels = defaultdict(set)

    def get(self, labels, key):
        """Get the key from the cache indicated by labels.

        This is the critical path. The rest of the methods need to accommodate for this
        to be as fast as possible.
        """
        # the real cache is under the tuple of labels
        cache = self._data.setdefault(labels, {})
        return cache.get(key)

    def set(self, labels, key, value):
        """Store the value under given key in the cache indicated by labels."""
        # store
        cache = self._data.setdefault(labels, {})
        cache[key] = value

        # also annotate the labels for future cleanup
        for label in labels:
            self._labels[label].add(labels)

    def clean(self, label):
        """Clean all caches for all set of labels where this label is present."""
        for section in self._labels[label]:
            self._data[section].clear()


class NvimNotifications:
    """Dance at the rhythm of Neovim.

    Hold the relevant structures and handle all notifications. Some are sent to the TextDisplay
    widgets, other to the main waindow.
    """

    def __init__(self, main_window):
        self.main_window = main_window
        self.text_display = None  # will be set before first usage
        self.options = {}

        # this two currently work "in tandem", we may want to unify them under the same structure
        # in the future
        self.structs = {}
        self.dyncache = DynamicCache()

    def _h__redraw(self, *parameters: tuple[Any]):
        """Handle the 'redraw' notification."""
        for submethod, *args in parameters:
            n_name = "_n_redraw__" + submethod
            n_meth = getattr(self, n_name, None)
            if n_meth is None:
                logger.error(
                    "[NvimNotifications] Submethod {!r} not implemented in 'redraw', params: {}",
                    submethod, args
                )
            else:
                try:
                    logger.debug("[NvimNotifications] Handle 'redraw': {} - {}", submethod, args)
                    n_meth(*args)
                except Exception:
                    logger.exception("Crash when calling {!r} with {!r}", n_name, args)

    def _h__modified_changed(self, is_modified: bool):
        """Handle the notification when the buffer starts/stop having changes."""
        self.main_window.set_buffer_state(is_modified=is_modified)

    def _h__filepath_changed(self, filepath: str):
        """Handle the notification when the buffer has a file associated or not."""
        self.main_window.set_buffer_state(filepath=filepath)

    def handler(self, method: str, parameters: list[Any]):
        """Handle all notifications from Neovim."""
        h_name = "_h__" + method
        h_meth = getattr(self, h_name, None)
        if h_meth is None:
            logger.error(
                "[NvimNotifications] Method {!r} not implemented, params: {}", method, parameters)
            return

        h_meth(*parameters)

    # -- specific notification handlers

    def _n_redraw__default_colors_set(self, colors):
        """Set the default colors."""
        rgb_fg, rgb_bg, rgb_sp, _, _ = colors  # last two are ignored because are for terminals
        self.structs.setdefault("default_colors", {}).update({
            "foreground": rgb_fg,
            "background": rgb_bg,
            "special": rgb_sp,
        })
        self.dyncache.clean("default_colors")

    def _n_redraw__flush(self, _):
        """Clear the grid."""
        self.text_display.flush()

    def _n_redraw__grid_clear(self, args):
        """Clear the grid."""
        (grid_id,) = args
        assert grid_id == 1  # FIXME.90: is it always 1? when do we have more than one?
        self.text_display.clear()

    def _n_redraw__grid_cursor_goto(self, args):
        """Resize a grid."""
        grid_id, row, col = args
        assert grid_id == 1  # FIXME.90: is it always 1? when do we have more than one?
        self.text_display.set_cursor(row, col)

    def _n_redraw__grid_line(self, *args):
        """Expose a line in the grid."""
        for item in args:
            grid, row, col_start, cells, wrap = item
            assert grid == 1  # FIXME.90: same question we do in grid_resize

            # note we ignore "wrap", couldn't find proper utility for it
            self.text_display.write_grid(row, col_start, cells)

    def _n_redraw__grid_resize(self, args):
        """Resize a grid."""
        grid_id, width, height = args
        assert grid_id == 1  # FIXME.90: is it always 1? when do we have more than one?
        self.text_display.resize_view((width, height))

    def _n_redraw__grid_scroll(self, args):
        """Scroll a grid."""
        grid_id, top, bottom, left, right, rows, cols = args
        assert grid_id == 1  # FIXME.90: is it always 1? when do we have more than one?
        self.text_display.scroll((top, bottom, rows), (left, right, cols))

    def _n_redraw__hl_attr_define(self, *args):
        """Add highlights with their attributes.

        E.g.: (
            2,
            {'foreground': 13882323, 'background': 11119017},
            {'foreground': 7, 'background': 242},
            [],
        )
        """
        hl_attrs = self.structs.setdefault("hl-attrs", {})
        for hl_id, rgb_attr, _, info in args:  # third value is ignored as it's for terminals
            assert not info
            hl_attrs[hl_id] = rgb_attr
        self.dyncache.clean("hl-attrs")

    def _n_redraw__hl_group_set(self, *args):
        """Set highlight groups.

        E.g.: [['SpecialKey', 161], ['EndOfBuffer', 161], ...]
        """
        hl_groups = self.structs.setdefault("hl-groups", {})
        for group_name, hl_id in args:
            hl_groups[group_name] = hl_id
        self.dyncache.clean("hl-groups")

    def _n_redraw__mode_change(self, args):
        """Information about cursor mode."""
        mode, mode_idx = args
        # we ignore the mode idx as we stored in the modes in a dict using the name
        mode_info = self.structs["mode-info"][mode]
        self.text_display.change_mode(mode_info)

    def _n_redraw__mode_info_set(self, args):
        """Information about cursor mode."""
        cursor_style_enabled, mode_info = args
        assert cursor_style_enabled  # may it come in False? what do we do? delete previous modes?

        info = {}
        for mi in mode_info:
            # store by name (and remove it from the real data, together with short name)
            name = mi.pop("name")
            del mi["short_name"]
            info[name] = mi

        self.structs.setdefault("mode-info", {}).update(info)
        self.dyncache.clean("mode-info")

    def _n_redraw__mouse_on(self, args):
        """Properly ignored."""

    def _n_redraw__mouse_off(self, args):
        """Properly ignored."""

    def _n_redraw__option_set(self, *options):
        options = dict(options)
        logger.debug("[NvimNotifications] options set: {}", options)
        self.options.update(options)

        # react to some of those options
        if "guifont" in options:
            name, size = options["guifont"].split(":")
            assert size[0] == "h"
            size = float(size[1:])
            self.text_display.set_font(name, size)

    def _n_redraw__set_icon(self, param):
        """Set the icon, if any."""
        (icon,) = param
        if icon:
            logger.warning("[NvimNotifications] need to implement set icon with {!r}", icon)

    def _n_redraw__set_title(self, param):
        """Set title."""
        (title,) = param
        self.main_window.setWindowTitle(title)

    def _n_redraw__win_viewport(self, args):
        """Information for the GUI viewport."""
        grid, objinfo, topline, botline, curline, curcol, line_count, scroll_delta = args
        # FIXME.90: ignore grid and objinfo so far, need to revisit this when multiwindow

        # Note: can't find use to scroll_delta (maybe for smooth scrollbar?)
        call_async(self.main_window.adjust_viewport, topline, botline, line_count, curcol)



# Copyright 2025-2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""Interface with the Neovim subprocess."""

import asyncio
import logging
import os
import subprocess
import socket
import time
import uuid

import msgpack

# do not poll more frequently than these seconds
POLL_FREEZE_PERIOD = 0.005

# some Neovim translation constants
_EXT_TYPE_CODES = {
    1: "window",
    2: "buffer",
    3: "tabpage",
}

logger = logging.getLogger(__name__)


class NeovimError(Exception):
    """Exception that indicates an error from Neovim."""
    def __init__(self, message=None):
        if message is None:
            message = self.__doc__
        super().__init__(message)


class NeovimExecutableNotFound(NeovimError):
    """The executable to run neovim was not found."""


def trace(msg, *params):
    """Log in trace level with a prefix."""
    logger.log(logging.TRACE, "[nvim] " + msg, *params)


def ext_hook(code, data):
    """Process other external types in msg-unpacking."""
    # code is the type of object
    obj_type = _EXT_TYPE_CODES[code]

    # Neovim encodes IDs as uint16 or uint32
    obj_id = int.from_bytes(data, byteorder='big')

    return [obj_type, obj_id]


class NvimInterface:
    """Interface with a Neovim process through RPC on a file socket."""

    def __init__(self, nvim_exec_path, loop, notification_handler, quit_callback):
        self._loop = loop
        self._notif_handler = notification_handler

        # the event is to wait for process finalization when we receive the order to close it; at
        # all times the callback is called, as the finalization may be initiated by Neovim itself
        self._quit_processed = asyncio.Event()
        self._quit_callback = quit_callback
        self._neovim_already_finished = False
        self._neovim_being_quited = False

        _sock_path = self._get_unique_sock_path()
        logger.info("Starting Neovim process, communicating through {!r}", _sock_path)
        if nvim_exec_path is None:
            nvim_exec_path = "nvim"
        try:
            self._proc = subprocess.Popen([nvim_exec_path, "--headless", "--listen", _sock_path])
        except FileNotFoundError as exc:
            logger.error("File not found when trying to run nvim: {!r}", exc)
            raise NeovimExecutableNotFound()

        tini = time.time()
        while True:
            if os.path.exists(_sock_path):
                break
            logger.debug("Wait nvim process to start")
            # yes, we block; if we want to go fully async here we need to find out a better
            # way to ensure that nvim is really up
            time.sleep(.05)
        tdelta = time.time() - tini
        logger.debug("Neovim process started! it took {:d} ms", int(tdelta / 1000))

        self._client = socket.socket(socket.AF_UNIX, socket.SOCK_STREAM)
        self._client.setblocking(False)
        self._client.connect(_sock_path)
        self._loop.add_reader(self._client, self._receive_responses)
        logger.info("Neovim setup done")

        self._callbacks = {}
        self._cb_counter = 0
        self._ui_attached = False
        self.last_poll_timestamp = 0

        self._msg_unpacker = msgpack.Unpacker(raw=False, ext_hook=ext_hook)

    def _get_unique_sock_path(self):
        """Return an unique path, validating it's not in disk."""
        while True:
            uniqueid = uuid.uuid4().hex
            path = f"/tmp/nysor-subnvim-{uniqueid}.s"
            if os.path.exists(path):
                logger.warning("Found an unique path already in disk: {!r}", path)
            else:
                return path

    async def quit(self):
        """Finish neovim and close the process.

        This is blocking on purpose, it normally should be fast but if it's not, it's a good idea
        to wait for it, maybe show more info at some point.

        Note that there is no real response to the 'quit' command, the process just disappears
        at some point.
        """
        if self._neovim_already_finished:
            return

        holder = [None]  # default, will remain like this if _quit_processed is set after ending OK

        def eback(message):
            """Error with the 'quit' command, pass the message."""
            holder[0] = message
            self._quit_processed.set()

        # this is a weird request; if all continues OK, Neovim will quit and a possible callback
        # is never called; however if there's a situation and Neovim can't quit, the errback
        # will be called
        self._neovim_being_quited = True
        await self._request(None, eback, "nvim_command", "quit")

        # sometimes the quit command needs an extra Enter, so send it!
        await self._request(None, None, "nvim_input", "\r")

        await self._quit_processed.wait()
        self._neovim_being_quited = False
        self._quit_processed.clear()  # for a possible next quitting attempt
        return holder[0]

    async def call(self, method, *params):
        """Call a method with the indicated parameters and wait until result is available."""
        call_result = None
        call_error = None
        event = asyncio.Event()

        def cback(result):
            nonlocal call_result

            call_result = result
            event.set()

        def eback(message):
            nonlocal call_error

            call_error = message
            event.set()

        await self._request(cback, eback, method, *params)
        await event.wait()
        if call_error:
            raise NeovimError(call_error)
        return call_result

    def future_request(self, method, *params):
        """Send a request in the future; response/error, if any, will be discarded."""
        self._loop.create_task(self._request(None, None, method, *params))

    async def _request(self, callback, errback, method, *params):
        """Send a 'request' message to run a method with some optional parameters.

        The callback will be executed when the information is available.
        """
        # filter out UI requests if UI still not attached
        if method == "nvim_ui_attach":
            self._ui_attached = True
        if method.startswith("nvim_ui") and not self._ui_attached:
            logger.debug("Ignoring request as UI still not attached; method: {!r}", method)

            # in case a callback is passed, just call it to close up waiters/coroutines; it's an
            # error condition, but the called method is doomed anyway
            if callback is not None:
                callback(None)
            return

        self._cb_counter += 1
        self._callbacks[self._cb_counter] = (callback, errback)

        # type (0 == request), msgid, method, params
        payload = msgpack.packb([0, self._cb_counter, method.encode("ascii"), params])
        trace("Sending request id={:d} method={!r} params={}", self._cb_counter, method, params)
        try:
            await self._loop.sock_sendall(self._client, payload)
        except BrokenPipeError:
            if not self._neovim_being_quited and not self._neovim_already_finished:
                raise

    def _read_messages(self):
        """Get messages from nvim."""
        while True:
            try:
                data = self._client.recv(4096)
            except BlockingIOError:
                data = ""
            if not data:
                break

            self._msg_unpacker.feed(data)
            for msg in self._msg_unpacker:
                yield msg

    def _receive_responses(self):
        """Receive responses from the nvim process; unpack, log, and send payloads to callbacks.

        Here is where we check if the process is still running, as it allows a frequent
        verification.
        """
        now = time.time()
        if (now - self.last_poll_timestamp) < POLL_FREEZE_PERIOD:
            return
        self.last_poll_timestamp = now

        # really read
        return_code = self._proc.poll()
        logger.debug("Reading response; rc {}", return_code)

        for msgtype, *rest in self._read_messages():
            if msgtype == 1:
                # response
                msgid, error, result = rest
                trace("Receiving response msgid={:d} error={!r} result={!r}", msgid, error, result)
                callback, errback = self._callbacks.pop(msgid)

                if error is not None:
                    logger.error("Error from Neovim: {!r}", error)
                    errback(error[1])
                if callback is not None:
                    callback(result)

            elif msgtype == 2:
                # notification
                method, params = rest
                trace("Receiving notification method={!r} params={!r}", method, params)
                self._notif_handler(method, params)

            else:
                logger.error("Bad message type from nvim: {!r} (rest={!r})", msgtype, rest)

        if return_code is not None:
            # process finished; do this closing at the end of the reading in any case there is a
            # final response to read
            logger.info("Neovim process finished")
            self._neovim_already_finished = True
            self._loop.remove_reader(self._client)
            self._quit_callback()
            self._quit_processed.set()  # 'quit' processed satisfactorily: Neovim ended



# Copyright 2025-2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""Init file to allow project's execution as a module."""

import sys

from nysor import main

sys.exit(main.start())



# Copyright 2025-2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""Main program."""

import argparse
import asyncio
import logging
import sys
import webbrowser

import qasync
from PyQt6.QtWidgets import (
    QFileDialog,
    QHBoxLayout,
    QMainWindow,
    QMessageBox,
    QScrollBar,
    QVBoxLayout,
    QWidget,
)
from PyQt6.QtCore import Qt
from PyQt6.QtGui import QIcon, QAction

from nysor.logtools import log_notdone, logsetup, LOG_LEVELS
from nysor.nvim_interface import NvimInterface, NeovimExecutableNotFound, NeovimError
from nysor.nvim_notifications import NvimNotifications
from nysor.text_display import TextDisplay

logger = logging.getLogger(__name__)


_NVIM_EXEC_NOT_FOUND_MSG = """\
<b>{exc}</b><br/>
<br/>
Find here how to install Neovim:<br/>
<br/>
&nbsp;&nbsp;&nbsp;&nbsp;<a href="https://neovim.io/doc/install/">https://neovim.io/doc/install/</a><br/>
<br/>
The <span style="font-family:monospace; color:green">nvim</span> executable should be in the system's PATH; alternatively you can indicate the path using the <span style="font-family:monospace; color:green">--nvim</span> parameter.
"""  # NOQA

_ABOUT_TEXT = """
<br/>
<span style="font-size:+1"><b>Nysor</b></span><br/>
<br/>
Yet another graphical interface for Neovim.<br/>
<br/>
Written in Python, with Qt.<br/>
<br/>
<br/>
<small>Copyright 2025-2026 Facundo Batista</small><br/>
"""


class MainMenu:
    """Build and manage the main menu bar for the application."""

    def __init__(self, main_window):
        self._main_window = main_window
        self._menu_bar = main_window.menuBar()
        self.actions = {}

        # FIXME.90 -- for multibuffers we need also a "New" option here
        menu_structure = {
            "&File": [
                ("&Open", "file__open"),
                ("&Save", "file__save"),
                ("&Save as...", "file__save_as"),
                (None, None),
                ("E&xit", "file__exit"),
            ],
            "&Debug": [
                ("Run a blocking call", "debug__blocking_call"),
                ("Run an async task", "debug__async_task"),
            ],
            "&Help": [
                ("Open &project page", "help__open_project_page"),
                ("Create a new &issue", "help__create_issue"),
                (None, None),
                ("&About Nysor", "help__about"),
            ],
        }

        for title, options in menu_structure.items():
            menu = self._menu_bar.addMenu(title)
            for visible_name, name in options:
                if visible_name is None:
                    menu.addSeparator()
                    continue

                action = QAction(visible_name, self._menu_bar)
                action.triggered.connect(getattr(self, f"_on__{name}"))
                self.actions[name] = action
                menu.addAction(action)

    def _log_action(func):
        """Log the action indicate by the user."""
        def _f(self):
            logger.debug("User triggered menu action {!r}", func.__name__)
            return func(self)
        return _f

    @_log_action
    def _on__file__open(self):
        """Open a file."""
        # FIXME.90 -- for multibuffers, logic will change here:
        # - if current buffer has any content, open the new file in a new panel
        # - if current is empty, replace it

        if self._main_window.state_buffer_is_modified:
            dlg = QMessageBox(self._main_window)
            dlg.setIcon(QMessageBox.Icon.Warning)
            dlg.setWindowTitle("Cannot open new file")
            dlg.setText(
                "Current buffer is not saved, cannot open a new file to replace it. "
                "Save the current buffer and try again."
            )
            dlg.setStandardButtons(QMessageBox.StandardButton.Ok)
            dlg.exec()
            return

        # indeed open a new file
        filename, _ = QFileDialog.getOpenFileName(self._main_window, "Open File", "", "")
        if filename:
            self._main_window.open_file(filename)

    @_log_action
    def _on__file__save(self):
        """Save the buffer in the current file."""
        if self._main_window.state_buffer_filepath:
            logger.debug("Saving the buffer directly with current name")
            self._main_window.buffer_save()
        else:
            filename, _ = QFileDialog.getSaveFileName(self._main_window, "Save File", "", "")
            if filename:
                logger.debug("Saving the buffer with new name {!r}", filename)
                self._main_window.save_to_new_file(filename)
            else:
                logger.debug("Saving the buffer cancelled, no new name chosen")

    @_log_action
    def _on__file__save_as(self):
        """Save the buffer in a new file."""
        filename, _ = QFileDialog.getSaveFileName(self._main_window, "Save File", "", "")
        if filename:
            logger.debug("Saving the buffer with new name {!r}", filename)
            self._main_window.save_to_new_file(filename)
        else:
            logger.debug("Saving the buffer cancelled, no new name chosen")

    @_log_action
    def _on__file__exit(self):
        """Exit the application."""
        self._main_window.close_gui()

    @_log_action
    def _on__debug__blocking_call(self):
        """Test an action triggered by the GUI; this is a test/dev helper."""
        logger.info("Code run in a standard function run from the GUI.")

    @_log_action
    def _on__debug__async_task(self):
        """Run an async task; this is a test/dev helper."""

        async def async_task():
            result = await self._main_window.nvi.call("nvim_list_uis")
            logger.info("Code run in an async task, listing UIs from Neovim: {}", result)

        asyncio.create_task(async_task())

    @_log_action
    def _on__help__open_project_page(self):
        """Open the project page in the browser."""
        webbrowser.open("https://github.com/facundobatista/nysor")

    @_log_action
    def _on__help__create_issue(self):
        """Open the issue tracker in the browser."""
        webbrowser.open("https://github.com/facundobatista/nysor/issues/new")

    @_log_action
    def _on__help__about(self):
        """Show the About dialog."""
        msg = _ABOUT_TEXT
        dlg = QMessageBox(self._main_window)
        dlg.setTextFormat(Qt.TextFormat.RichText)
        dlg.setIconPixmap(QIcon("nysor/imgs/icon-1024.png").pixmap(128, 128))
        dlg.setWindowTitle("About Nysor")
        dlg.setText(msg)
        dlg.setStandardButtons(QMessageBox.StandardButton.Ok)
        dlg.exec()


class MainApp(QMainWindow):
    """The main application window."""

    def __init__(self, loop, path_to_open, nvim_exec_path):
        super().__init__()
        self.setWindowIcon(QIcon("nysor/imgs/icon-1024.png"))
        self._menu = MainMenu(self)

        logger.info("Starting Nysor")
        self._closing = 0
        self.state_buffer_is_modified = False
        self.state_buffer_filepath = None
        self.nvim_notifs = NvimNotifications(self)

        # setup the Neovim interface
        try:
            self.nvi = NvimInterface(
                nvim_exec_path, loop, self.nvim_notifs.handler, self._quit_callback
            )
        except NeovimExecutableNotFound as exc:
            logger.error("Failed to start neovim interface: {!r}", exc)
            msg = _NVIM_EXEC_NOT_FOUND_MSG.format(exc=exc)
            dlg = QMessageBox(self)
            dlg.setTextFormat(Qt.TextFormat.RichText)
            dlg.setIcon(QMessageBox.Icon.Critical)
            dlg.setWindowTitle("Startup Error")
            dlg.setText(msg)
            dlg.setStandardButtons(QMessageBox.StandardButton.Ok)
            dlg.exec()
            exit(1)

        loop.create_task(self.setup_nvim(path_to_open))

        # scrollbars
        self.v_scroll = QScrollBar(Qt.Orientation.Vertical)
        self.v_scroll.valueChanged.connect(self.vertical_scroll_changed)
        self.v_scroll.setMinimum(0)
        self.v_scroll.setMaximum(100)
        self.v_scroll_last_position = None
        self.h_scroll = QScrollBar(Qt.Orientation.Horizontal)
        self.h_scroll.valueChanged.connect(self.horizontal_scroll_changed)
        self.h_scroll.setMinimum(0)
        self.h_scroll.setMaximum(100)
        self.h_scroll_last_position = None

        # central widget to hold main layout
        self.central_widget = QWidget(self)
        self.setCentralWidget(self.central_widget)
        self.main_layout = QVBoxLayout(self.central_widget)

        # horizontal layout for text display and vertical scroll bar
        hbox = QHBoxLayout()
        self.text_display = self.nvim_notifs.text_display = TextDisplay(self)
        self.text_display.setFocus()
        hbox.addWidget(self.text_display, stretch=1)
        hbox.addWidget(self.v_scroll)
        self.main_layout.addLayout(hbox)

        # rest of vertical widgets
        self.main_layout.addWidget(self.h_scroll)

    def set_buffer_state(self, is_modified=None, filepath=None):
        """Set the state state."""
        # FIXME.90 -- all these needs to evolve to multibuffers
        if is_modified is not None:
            self.state_buffer_is_modified = is_modified
            self._menu.actions["file__save"].setEnabled(is_modified)
            self._menu.actions["file__open"].setEnabled(not is_modified)
        if filepath is not None:
            self.state_buffer_filepath = filepath

    async def setup_nvim(self, path_to_open):
        """Set up Neovim."""
        channel_id, api_metadata = await self.nvi.call("nvim_get_api_info")
        version = api_metadata["version"]
        major = version["major"]
        minor = version["minor"]
        patch = version["patch"]
        logger.info("Neovim API info: version {}.{}.{}", major, minor, patch)

        # attach the UI
        nvim_config = {"ext_linegrid": True}
        await self.nvi.call("nvim_ui_attach", 80, 20, nvim_config)

        # subscribe to the buffer file change (will comeback as a 'set_buffer_state' call)
        # FIXME.90 -- this needs to evolve to multibuffers
        _code = f"""
            vim.api.nvim_create_autocmd({{'BufFilePost'}}, {{
                callback = function()
                    local name = vim.api.nvim_buf_get_name(0)
                    vim.rpcnotify({channel_id}, 'filepath_changed', name)
                end
            }})
        """
        await self.nvi.call("nvim_exec_lua", _code, [])

        # subscribe to the buffer modified change (will comeback as a 'set_buffer_state' call)
        # FIXME.90 -- this needs to evolve to multibuffers
        _code = f"""
            vim.api.nvim_create_autocmd({{'BufModifiedSet', 'BufWritePost'}}, {{
                callback = function()
                    vim.rpcnotify({channel_id}, 'modified_changed', vim.bo.modified)
                end
            }})
        """
        await self.nvi.call("nvim_exec_lua", _code, [])
        self.set_buffer_state(is_modified=False)  # initially it's always not modified

        if path_to_open is not None:
            # FIXME.92: properly support "-" to read from stdin
            cmd = {"cmd": "edit", "args": [path_to_open]}
            opts = {"output": False}  # don't capture output
            try:
                await self.nvi.call("nvim_cmd", cmd, opts)
            except NeovimError as err:
                log_notdone("Got error when opening the file {}, this should never happen", err)

    async def _quit(self):
        """Close the GUI after Neovim is down."""
        logger.debug("Start shutdown, asking")
        error = await self.nvi.quit()
        if error:
            dlg = QMessageBox(self)
            dlg.setIcon(QMessageBox.Icon.Warning)
            dlg.setWindowTitle("Neovim Error")
            dlg.setText(error)
            dlg.setStandardButtons(QMessageBox.StandardButton.Ok)
            dlg.exec()
            self._closing = 0  # reset
            return

        self._closing = 2  # allows final close
        logger.debug("Start shutdown, done")
        self.close()

    def _quit_callback(self):
        """Close the GUI because of nvim interface request."""
        if self._closing == 0:
            # only if it was not initiated internally
            logger.debug("Shutdown requested by nvim interface")
            self._closing = 2
            self.close()

    def close_gui(self, event=None):
        """Close Neovim and then let the rest to finish.

        If event is not None this is the closeEvent (from alt-F4, click in the X, etc.), so
        ignoring/accepting it is the real indication to close the GUI.

        If event is None (called from the menu) we need to call 'close' explicitly.
        """
        logger.debug("Close requested (event={}); current state {:d}", event, self._closing)
        if self._closing == 0:
            # initiate the process to close; ignore the event so GUI is still alive, but
            # start internal procedures
            self._closing = 1
            if event is not None:
                event.ignore()
            asyncio.create_task(self._quit())
            return

        if self._closing == 1:
            # the request was received again in the middle of internal closing, keep ignoring it
            if event is not None:
                event.ignore()
            return

        # we're really done here, let the event propagate if we have it, or close manually
        assert self._closing == 2
        logger.debug("Bye")
        if event is None:
            self.close()
        else:
            event.accept()

    def closeEvent(self, event):
        """Handle the event from the GUI to close itself."""
        self.close_gui(event)

    def present_context_window(self):
        """Present a context window with some options for the user."""
        # FIXME.93
        log_notdone("Mouse context window!")

    async def adjust_viewport(self, topline, botline, line_count, curcol):
        """Adjust scrollbar according to what Neovim says.

        Called from the Neovim layer on viewport changes notifications.
        """
        display_width, display_height = self.text_display.display_size

        # vertical: use information from the viewport
        if topline == 0 and line_count <= display_height:
            self.v_scroll.setEnabled(False)
            self.v_scroll.setMaximum(0)
        else:
            self.v_scroll.setEnabled(True)
            self.v_scroll.setPageStep(display_height)
            self.v_scroll.setMaximum(line_count - 1)
            self.v_scroll_last_position = topline  # before setting value to ignore later event
            self.v_scroll.setValue(topline)

        # only if not wrapping, using current column but also queried line lengths
        # FIXME.90: a "scope" can be given here, revisit this when we have multiple windows
        is_wrapping = await self.nvi.call("nvim_get_option_value", "wrap", {})
        if is_wrapping:
            # just turn off the scroll bar as when wrapping all text will be inside the window
            self.h_scroll.setEnabled(False)
            self.h_scroll.setMaximum(0)
            return

        # get the lengths of lines that are currently shown
        buf = 0  # FIXME.90: why 0? first buffer per window? revisit when multiple windows
        start = topline + 1  # getbufline's first line is 1
        end = botline - 1  # botline is the "next line, out of the view"
        cmd = f"map(getbufline({buf}, {start}, {end}), {{key, val -> strlen(val)}})"
        line_lengths = await self.nvi.call("nvim_eval", cmd)
        if not line_lengths:
            return

        max_line = max(line_lengths)
        if max_line <= display_width:
            self.h_scroll.setEnabled(False)
            self.h_scroll.setMaximum(0)
        else:
            self.h_scroll.setEnabled(True)
            self.h_scroll.setPageStep(display_width)
            self.h_scroll.setMaximum(max_line - 1)

            win_info = await self.nvi.call("nvim_eval", "winsaveview()")
            leftcol = win_info["leftcol"]
            self.h_scroll_last_position = leftcol  # before setting value to ignore later event
            self.h_scroll.setValue(leftcol)

    def vertical_scroll_changed(self, value):
        """Handle the vertical scroll bar being modified through the widget."""
        delta = value - self.v_scroll_last_position
        if delta > 0:
            # down
            cmdkey = "\x05"
        elif delta < 0:
            # up
            cmdkey = "\x19"
        else:
            return
        self.v_scroll_last_position = value
        self.nvi.future_request("nvim_command", f"normal! {abs(delta)}{cmdkey}")

    def horizontal_scroll_changed(self, value):
        """Handle the horizontal scroll bar being modified through the widget."""
        delta = value - self.h_scroll_last_position
        if delta > 0:
            # right
            cmdkey = "zl"
        elif delta < 0:
            # left
            cmdkey = "zh"
        else:
            return
        self.h_scroll_last_position = value
        self.nvi.future_request("nvim_command", f"normal! {abs(delta)}{cmdkey}")

    # -- set of functions to interact with buffers/neovim

    def open_file(self, filename):
        """Tell neovim to open a file and load into current buffer."""
        self.nvi.future_request("nvim_command", f"edit {filename}")

    def buffer_save(self):
        """Save the current buffer."""
        self.nvi.future_request("nvim_command", "write")

    def save_to_new_file(self, filename):
        """Save current buffer to a new filename."""
        self.nvi.future_request("nvim_command", f"saveas {filename}")


def start():
    """Start the application."""
    # mutually exclusive verbosity levels
    parser = argparse.ArgumentParser()
    loggroup = parser.add_mutually_exclusive_group()
    for option, (_, helpmsg) in LOG_LEVELS.items():
        if option:
            loggroup.add_argument(
                f"-{option[0]}",
                f"--{option}",
                action="store_const",
                const=option,
                dest="loglevel",
                help=helpmsg
            )

    # the rest of argument parsing
    parser.add_argument("--nvim", action="store", help="Path to the Neovim executable.")
    parser.add_argument(
        "path", action="store", nargs="?", default=None,
        help="Path to the file to edit or directory to open (optional)"
    )

    # parse arguments
    args = parser.parse_args()

    # setup logging and create the app itself
    logsetup(args.loglevel)
    app = qasync.QApplication(sys.argv)

    # connect with async's event loop
    event_loop = qasync.QEventLoop(app)
    event_loop.set_debug(True)
    asyncio.set_event_loop(event_loop)
    app_close_event = asyncio.Event()
    app.aboutToQuit.connect(app_close_event.set)

    # start and show GUI
    main_window = MainApp(event_loop, args.path, args.nvim)
    main_window.show()

    # go!
    with event_loop:
        event_loop.run_until_complete(app_close_event.wait())



# Copyright 2025-2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""Logical lines."""

import logging
from dataclasses import dataclass

from PyQt6.QtGui import QColor

logger = logging.getLogger(__name__)


@dataclass
class CharUnderline:
    """Represent the underlyne style of a single char."""
    color: QColor
    style: str  # "underline", "undercurl", "underdouble", "underdotted", "underdashed"


@dataclass
class CharFormat:
    """Represent the overall format of a single char."""
    foreground: QColor
    background: QColor
    strikethrough: bool = False
    italic: bool = False
    bold: bool = False
    underline: CharUnderline | None = None


@dataclass
class LogicalChar:
    """Represent a logical character with its format and wide indication."""

    char: str
    format: CharFormat
    is_wide: bool = False

    def __hash__(self):
        return hash((self.char, self.is_wide))

    def __eq__(self, other):
        return (self.char, self.is_wide) == (other.char, other.is_wide)


class LogicalLines:
    """Hold the lines to show in the grid."""

    def __init__(self, q_rows, q_cols, fmt):
        self._lines = {
            idx: [LogicalChar(" ", fmt) for _ in range(q_cols)] for idx in range(q_rows)
        }

    @classmethod
    def empty(cls):
        """Return an empty line."""
        return cls(0, 0, None)

    def get(self, row):
        """Return the logical line for the indicated row."""
        return self._lines.get(row)

    def add(self, row, col, textinfo):
        """Add text info to the grid."""
        # expand the textinfo so we have one format per character, to keep our logical grid
        expanded = []
        for text, fmt in textinfo:
            if text is None:
                # special "char" that comes after others to indicate those are width: we flag
                # the previous item and keep the position with None for the slice assignment
                # below to work correctly
                expanded[-1].is_wide = True
                expanded.append(None)
            else:
                expanded.extend(LogicalChar(char, fmt) for char in text)

        # it's fine to create new lines in the map, because gaps are created when scrolling
        prvline = self._lines.setdefault(row, [])

        if col > len(prvline):
            raise ValueError("Trying to write outside the line; needs to rethink model!!!")
        prvline[col: col + len(expanded)] = expanded

    def scroll_vertical(self, top, bottom, delta):
        """Scroll vertically some lines in the grid."""
        if abs(delta) >= bottom - top:
            raise ValueError("Overflow scrolling; not supported yet")

        if delta > 0:
            # goes up; move N lines up and then remove "hole"
            for idx in range(top + delta, bottom):
                self._lines[idx - delta] = self._lines[idx]
            for idx in range(bottom - delta, bottom):
                del self._lines[idx]

        elif delta < 0:
            # goes down; move N lines down and then remove "hole"
            for idx in reversed(range(top, bottom + delta)):
                self._lines[idx - delta] = self._lines[idx]
            for idx in range(top, top - delta):
                del self._lines[idx]

        else:
            logger.warning("Called scroll vertical with delta=0, shouldn't happen")
            return



# Copyright 2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""Small utilities to help other code."""

import asyncio
import logging

logger = logging.getLogger(__name__)


# collection so futurized tasks are always referenced while alive, to avoid
# premature garbage collection
_futurized_background_tasks = set()


def _future_cleanup(task):
    """Cleanup for futurized tasks."""
    # remove it from the global holding set
    _futurized_background_tasks.discard(task)

    # get the result only to properly log an exception if happened
    try:
        task.result()
    except Exception:
        logger.exception("Futurized call crashed")


def call_async(async_function, *args, **kwargs):
    """Call an async function / couroutine in the future.

    Used to call coroutines from blocking code. As the execution is in the future, there is no
    way to get the result of that call. This function just returns None.
    """
    # get the coroutine, and a task from it
    coro = async_function(*args, **kwargs)
    task = asyncio.create_task(coro)

    # include it in the global set so it's never un-referenced until completion
    _futurized_background_tasks.add(task)

    # plan the cleanup
    task.add_done_callback(_future_cleanup)



# Copyright 2025-2026 Facundo Batista
# Licensed under the Apache v2 License
# For further info, check https://github.com/facundobatista/nysor

"""The widget that display all the text from Neovim."""

import logging
import math
from dataclasses import dataclass
from typing import Any
from functools import partial

from PyQt6.QtWidgets import QWidget
from PyQt6.QtGui import (
    QColor,
    QFont,
    QFontMetricsF,
    QKeyEvent,
    QMouseEvent,
    QPaintEvent,
    QPainter,
    QPainterPath,
    QPen,
    QResizeEvent,
    QWheelEvent,
)
from PyQt6.QtCore import QPointF, Qt, QRectF, QSize

from nysor.logical_lines import LogicalLines, CharFormat, CharUnderline
from nysor.logtools import log_notdone

logger = logging.getLogger(__name__)


@dataclass
class FontSize:
    """Hold information about cell and char inside it."""
    width: float
    height: float
    ascent: float


# conversion between the names used by Neovim to the styles defined in Qt6
UNDERLINE_STYLES = [
    "underline",
    "undercurl",
    "underdouble",
    "underdotted",
    "underdashed",
]

# never ask to Neovim a grid smaller than these
MIN_COLS_ROWS = 5

# conversion between Qt key codes and Neovim names for some special keys
QT_NVIM_KEYS_MAP = {
    Qt.Key.Key_Left: "Left",
    Qt.Key.Key_Right: "Right",
    Qt.Key.Key_Up: "Up",
    Qt.Key.Key_Down: "Down",
    Qt.Key.Key_Home: "Home",
    Qt.Key.Key_End: "End",
    Qt.Key.Key_PageUp: "PageUp",
    Qt.Key.Key_PageDown: "PageDown",
    Qt.Key.Key_Insert: "Insert",
    Qt.Key.Key_Delete: "Del",
    Qt.Key.Key_Backspace: "BS",
    Qt.Key.Key_Return: "CR",
    Qt.Key.Key_Enter: "Enter",
    Qt.Key.Key_Tab: "Tab",
    Qt.Key.Key_Escape: "Esc",
    Qt.Key.Key_F1: "F1",
    Qt.Key.Key_F2: "F2",
    Qt.Key.Key_F3: "F3",
    Qt.Key.Key_F4: "F4",
    Qt.Key.Key_F5: "F5",
    Qt.Key.Key_F6: "F6",
    Qt.Key.Key_F7: "F7",
    Qt.Key.Key_F8: "F8",
    Qt.Key.Key_F9: "F9",
    Qt.Key.Key_F10: "F10",
    Qt.Key.Key_F11: "F11",
    Qt.Key.Key_F12: "F12",
}

# Handier
MouseButton = Qt.MouseButton


class BaseDisplay(QWidget):
    """Base widget to isolate as much as possible Qt itself from the Text handling."""

    def __init__(self):
        super().__init__()
        self.widget_size = QSize(100, 100)  # default valid pseudo-useful value
        self.setMouseTracking(True)

        # get *all* keyboard events in this widget
        self.setFocusPolicy(Qt.FocusPolicy.StrongFocus)

    def focusNextPrevChild(self, _):
        """Do not allow to "navigate" widgets out of here."""
        return False

    def sizeHint(self):
        """Provide the desired size for the widget."""
        return self.widget_size

    def resizeEvent(self, event: QResizeEvent):
        """Hook-up in the event to trigger internal resizing."""
        super().resizeEvent(event)
        self.window_resize()

    def keyPressEvent(self, event: QKeyEvent):
        """Get all keyboard events."""
        key_text = event.text()
        key = event.key()
        modifiers = event.modifiers()
        self.handle_keyboard(key_text, key, modifiers)

    def paintEvent(self, event: QPaintEvent):
        """Paint the widget."""
        painter = QPainter(self)
        painter.setRenderHints(
            QPainter.RenderHint.Antialiasing | QPainter.RenderHint.TextAntialiasing
        )
        self.paint(painter)
        painter.end()

    # ----- set of mouse related callbacks
    #
    # Nothing is really done by the TextDisplay, but by Neovim directly or the main window:
    #
    # left button:
    #    - on press: sent to Neovim, for cursor positioning
    #    - on release: same, but also some chars are sent to handle clipboard properly; note that
    #        double and triple clicking, and related selections, is automatically handled by Neovim
    #
    #  right button:
    #    - on release: called a function in main window to present a context window
    #
    #  middle button:
    #  - on press: sent to Neovim, for clipboard pasting
    #
    #  wheel: sent to Neovim, for displacement
    #
    #  mouse movement: only sent to Neovim if left button is pressed, for "dragging", implies
    #    selection of text
    #
    #  notes:
    #    - in every event sent to Neovim, whatever modifier is used is also sent
    #    - if button is not left/middle/right, is considered left
    #    - if an event is not described above, it's ignored
    #

    def mousePressEvent(self, event: QMouseEvent):
        """Handle a button mouse that was pressed."""
        button = event.button()

        if button is MouseButton.RightButton:
            # ignored
            return

        if button is MouseButton.MiddleButton:
            button_name = "middle"
        else:
            # really left, or default to left as mouses can have a ton of buttons
            button_name = "left"

        action = "press"
        modifier = self._get_button_modifiers(event)
        grid = 0  # FIXME.90: may change when multi-edit?

        pos = event.position()
        row, col = self._get_grid_cell(pos.x(), pos.y())
        self.main_window.nvi.future_request(
            "nvim_input_mouse", button_name, action, modifier, grid, row, col
        )

    def mouseReleaseEvent(self, event: QMouseEvent):
        """Handle a button mouse that was released."""
        button = event.button()
        if button is MouseButton.RightButton:
            self.main_window.present_context_window()
            return

        if button is MouseButton.MiddleButton:
            # ignored
            return

        # really left, or default to left as mouses can have a ton of buttons
        button_name = "left"

        action = "release"
        modifier = self._get_button_modifiers(event)
        grid = 0  # FIXME.90: may change when multi-edit?

        pos = event.position()
        row, col = self._get_grid_cell(pos.x(), pos.y())
        self.main_window.nvi.future_request(
            "nvim_input_mouse", button_name, action, modifier, grid, row, col
        )

        # this will make Neovim to yank selection to the "X11 main selection"
        self.main_window.nvi.future_request("nvim_command", 'normal! "*ygv')

    def mouseMoveEvent(self, event: QMouseEvent):
        """Handle when the mouse is moving; we only care about this for left button dragging."""
        button = event.buttons()
        if button in (MouseButton.NoButton, MouseButton.RightButton, MouseButton.MiddleButton):
            # ignore the event if not dragging with left button
            return

        # really left, or default to left as mouses can have a ton of buttons
        button_name = "left"
        action = "drag"
        modifier = self._get_button_modifiers(event)
        grid = 0  # FIXME.90: may change when multi-edit?

        pos = event.position()
        row, col = self._get_grid_cell(pos.x(), pos.y())
        self.main_window.nvi.future_request(
            "nvim_input_mouse", button_name, action, modifier, grid, row, col
        )

    def wheelEvent(self, event: QWheelEvent):
        """Handle when the wheel is used (or a wheel like interface).

        The event information is an angle. As Neovim only supports to be informed in one
        direction, in case of having both displacements in X and Y, we inform twice.

        Also, we inform only when the displacement is noticeable, for interface
        not to be "too nervous". Note that typical wheel of standard mouses will
        inform a delta of 120.
        """
        button_name = "wheel"
        qpoint = event.angleDelta()
        dx, dy = qpoint.x(), qpoint.y()
        trigger_limit = 10
        row, col = 0, 0  # seems to be ignored
        grid = 0  # FIXME.90: may change when multi-edit?

        if abs(dx) > trigger_limit:
            action = "right" if dx > 0 else "left"
            modifier = self._get_button_modifiers(event)
            self.main_window.nvi.future_request(
                "nvim_input_mouse", button_name, action, modifier, grid, row, col
            )

        if abs(dy) > trigger_limit:
            action = "up" if dy > 0 else "down"
            modifier = self._get_button_modifiers(event)
            self.main_window.nvi.future_request(
                "nvim_input_mouse", button_name, action, modifier, grid, row, col
            )

    def _get_grid_cell(self, x: int, y: int):
        """Return grid's row and column from pixels x and y."""
        col = int(x / self.font_size.width)
        row = int(y / self.font_size.height)
        return row, col

    def _get_button_modifiers(self, event: QMouseEvent | QWheelEvent):
        """Return a string indicating the used modifiers, to inform Neovim."""
        indicators = []
        mods = event.modifiers()
        if mods & Qt.KeyboardModifier.ShiftModifier:
            indicators.append("S")
        if mods & Qt.KeyboardModifier.ControlModifier:
            indicators.append("C")
        if mods & Qt.KeyboardModifier.AltModifier:
            indicators.append("A")
        return "-".join(indicators)

    # ----- end of mouse related event handling methods


class TextDisplay(BaseDisplay):
    """A text display widget."""

    # cache to hold chars drawing widths; cleaned when font changes
    _char_drawing_widths_cache = {}

    def __init__(self, main_window):
        super().__init__()
        self.main_window = main_window
        self.initial_resizing_done = False

        # some defaults
        self.font_size = None
        self.display_size = (80, 20)
        self.set_font("Courier", 12)

        self.lines = LogicalLines.empty()
        self.cursor_pos = (0, 0)
        self.cursor_painter = lambda *a: None
        self.need_grid_clearing = True

        # cache to hold conversions between Neovim's highlight info and Qt formats
        self.nvimhl_to_qtfmt = {}
        # cache to hold mode_info processed structures
        self.mode_info_structs = {}

    def window_resize(self):
        """Inform Neovim of new window size."""
        cols = max(MIN_COLS_ROWS, int(self.width() / self.font_size.width))
        rows = max(MIN_COLS_ROWS, int(self.height() / self.font_size.height))
        self.main_window.nvi.future_request("nvim_ui_try_resize", cols, rows)

    def handle_keyboard(self, key_text, key, modifiers):
        """Handle keyboard events."""
        # simple case when it's just unicode text
        if key_text:
            key_text = key_text.replace("<", "<LT>")
            self.main_window.nvi.future_request("nvim_input", key_text)
            return

        # need to compose special keys
        keyname = QT_NVIM_KEYS_MAP.get(key)
        if keyname is None:
            return

        parts = []
        if modifiers & Qt.KeyboardModifier.ControlModifier:
            parts.append("C")
        if modifiers & Qt.KeyboardModifier.ShiftModifier:
            parts.append("S")
        if modifiers & Qt.KeyboardModifier.AltModifier:
            parts.append("A")
        if modifiers & Qt.KeyboardModifier.MetaModifier:
            parts.append("D")  # 'D' often represents 'Command' in Mac

        parts.append(keyname)
        composed = f"<{"-".join(parts)}>"
        self.main_window.nvi.future_request("nvim_input", composed)

    def _build_empty_logical_lines(self):
        """Build an empty logical lines."""
        default_fmt = self._build_text_format(None)
        cols, rows = self.display_size
        return LogicalLines(rows, cols, default_fmt)

    def clear(self):
        """Clear the display."""
        self.need_grid_clearing = True
        self.lines = self._build_empty_logical_lines()

    def set_font(self, name, size):
        """Set the font."""
        # when requesting the font itself, round up the size, as it may not work
        # properly with non-ints
        self.font = QFont(name, math.ceil(size))
        self.font.setFixedPitch(True)

        # try to set the real size, however it may not work in all systems
        self.font.setPointSizeF(size)

        # clear the cache for the drawing widths
        self._char_drawing_widths_cache.clear()

        # store font sizes
        fm = QFontMetricsF(self.font)
        char_width = fm.horizontalAdvance("M")
        line_height = fm.height()
        self.font_size = FontSize(width=char_width, height=line_height, ascent=fm.ascent())
        self.resize_view(force=True)

    def resize_view(self, size=None, force=False):
        """Resize the display.

        If size is given (W x H) it is used; else use current size (if not set, default to 80x20.
        """
        if size is None:
            size = self.display_size
        else:
            self.display_size = size
        cols, rows = size

        # adjust display size for font, if we have it
        if self.font_size is None:
            return

        # calculate widget desired size, update internal record, and call Qt magic to resize/redraw
        view_width = math.ceil(self.font_size.width * cols)
        view_height = math.ceil(self.font_size.height * rows)
        self.widget_size = QSize(view_width, view_height)
        if force:
            self.updateGeometry()
            self.main_window.adjustSize()

    def scroll(self, vertical, horizontal):
        """Scroll the grid."""
        top, bottom, delta = vertical
        if delta:
            self.lines.scroll_vertical(top, bottom, delta)

        left, right, delta = horizontal
        if delta != 0:
            log_notdone("Horizontal scroll has delta", delta=delta)

    def flush(self):
        """Update the window."""
        self.update()

    def set_cursor(self, row, col):
        """Set the cursor position in the display."""
        self.cursor_pos = (row, col)

    def _paint_cursor_block(self, attr_id, painter, start_x, start_y, width):
        """Draw a cursor as a block."""
        rect = QRectF(start_x, start_y, width, self.font_size.height)
        assert attr_id == 0  # means inverting color, which is what we're only doing here
        painter.save()
        painter.setCompositionMode(QPainter.CompositionMode.RasterOp_SourceXorDestination)
        painter.fillRect(rect, Qt.GlobalColor.white)
        painter.restore()

    def _paint_cursor_vertical(self, attr_id, percentage, painter, start_x, start_y, width):
        """Draw a cursor as a vertical line."""
        rect = QRectF(start_x, start_y, width * percentage / 100, self.font_size.height)
        assert attr_id == 0  # means inverting color, which is what we're only doing here
        painter.save()
        painter.setCompositionMode(QPainter.CompositionMode.RasterOp_SourceXorDestination)
        painter.fillRect(rect, Qt.GlobalColor.white)
        painter.restore()

    def _paint_cursor_horizontal(self, attr_id, percentage, painter, start_x, start_y, width):
        """Draw a cursor as an horizontal line."""
        from_y = start_y + self.font_size.height * (1 - percentage / 100)
        height = self.font_size.height * percentage / 100
        rect = QRectF(start_x, from_y, width, height)
        assert attr_id == 0  # means inverting color, which is what we're only doing here
        painter.save()
        painter.setCompositionMode(QPainter.CompositionMode.RasterOp_SourceXorDestination)
        painter.fillRect(rect, Qt.GlobalColor.white)
        painter.restore()

    def _get_drawing_widths(self, logical_char):
        """Define the values for placing chars in the line.

        This is cached per char; note this cache is cleaned when font changes.

        Returns the slot and char widths, and the horizontal shift to start drawing.
        """
        try:
            return self._char_drawing_widths_cache[logical_char]
        except KeyError:
            # not in the cache: calculate, store, and return values
            pass

        fm = QFontMetricsF(self.font)
        char_width = fm.horizontalAdvance(logical_char.char)

        slot_width = self.font_size.width
        shift = 0
        if logical_char.is_wide:
            slot_width *= 2
            shift = (slot_width - char_width) / 2

        values = slot_width, shift, char_width
        self._char_drawing_widths_cache[logical_char] = values
        return values

    def paint(self, painter):
        """Paint (draw) the grid."""
        cell_height = self.font_size.height

        # paint all backgrounds first!
        for row in range(self.display_size[1]):
            base_y = row * cell_height
            base_x = 0

            logical_line = self.lines.get(row)
            if logical_line is None:
                # no logical line, fill with background default color; note that this value is not
                # ready at the very start, but it's there soon enough
                rect = QRectF(0, base_y, self.width(), cell_height)
                default_colors = self.main_window.nvim_notifs.structs.get("default_colors")
                if default_colors is not None:
                    painter.fillRect(rect, QColor(default_colors["background"]))
                continue

            for logical_char in logical_line:
                if logical_char is None:
                    continue

                slot_width, _, _ = self._get_drawing_widths(logical_char)
                x = base_x
                base_x += slot_width

                rect = QRectF(x, base_y, slot_width + 1, cell_height)
                painter.fillRect(rect, logical_char.format.background)

        # the foregrounds
        cursor_row, cursor_col = self.cursor_pos
        for row in range(self.display_size[1]):
            base_y = row * cell_height
            base_x = 0

            logical_line = self.lines.get(row)
            if logical_line is None:
                continue

            for col, logical_char in enumerate(logical_line):
                if logical_char is None:
                    continue

                # get the value for current x, and shift the base for next round
                slot_width, horizontal_shift, char_width = self._get_drawing_widths(logical_char)

                # base font
                self.font.setItalic(logical_char.format.italic)
                self.font.setBold(logical_char.format.bold)
                painter.setFont(self.font)

                # foreground
                painter.setPen(logical_char.format.foreground)

                # draw the text
                text_x = base_x + horizontal_shift
                text_y = base_y + (cell_height + self.font_size.ascent) / 2 - 2
                painter.drawText(QPointF(text_x, text_y), logical_char.char)

                # and effects over the test
                if logical_char.format.strikethrough:
                    self._draw_strikethrough(painter, logical_char, text_x, char_width, text_y)
                if logical_char.format.underline:
                    self._draw_underline(
                        painter, logical_char, base_x, slot_width, base_y, cell_height,
                    )

                # the cursor, if that is the position
                if col == cursor_col and row == cursor_row:
                    self.cursor_painter(painter, base_x, base_y, slot_width - 1)

                base_x += slot_width

    def _draw_strikethrough(self, painter, logical_char, text_x, char_width, text_y):
        """Draw strikethrough effect over the text."""
        painter.setPen(logical_char.format.foreground)
        strike_y = text_y - self.font_size.ascent / 3
        rect = QRectF(text_x, strike_y, text_x + char_width, strike_y)
        painter.drawLine(rect)

    def _draw_underline(self, painter, logical_char, base_x, slot_width, base_y, cell_height):
        """Draw underline effect over the text."""
        underline_y = int(base_y + cell_height - 1)

        match logical_char.format.underline.style:
            case "underline":
                pen = QPen(logical_char.format.underline.color)
                pen.setWidth(2)
                painter.setPen(pen)
                painter.drawLine(
                    QPointF(base_x, underline_y),
                    QPointF(base_x + slot_width, underline_y)
                )

            case "underdotted":
                pen = QPen(logical_char.format.underline.color)
                pen.setStyle(Qt.PenStyle.DotLine)
                painter.setPen(pen)
                painter.drawLine(
                    QPointF(base_x, underline_y),
                    QPointF(base_x + slot_width, underline_y)
                )

            case "underdashed":
                pen = QPen(logical_char.format.underline.color)
                pen.setStyle(Qt.PenStyle.DashLine)
                painter.setPen(pen)
                painter.drawLine(
                    QPointF(base_x, underline_y),
                    QPointF(base_x + slot_width, underline_y)
                )

            case "underdouble":
                pen = QPen(logical_char.format.underline.color)
                painter.setPen(pen)
                painter.drawLine(
                    QPointF(base_x, underline_y),
                    QPointF(base_x + slot_width, underline_y)
                )
                painter.drawLine(
                    QPointF(base_x, underline_y + 3),
                    QPointF(base_x + slot_width, underline_y + 3)
                )

            case "undercurl":
                path = QPainterPath()
                amplitude = 1
                period = 4
                underline_y += 1
                path.moveTo(base_x, underline_y)
                i = 0
                while i < slot_width:
                    cx1 = base_x + i + period / 2
                    cy1 = underline_y + (amplitude if (i // period) % 2 == 0 else -amplitude)
                    path.lineTo(cx1, cy1)
                    i += period
                painter.setPen(QPen(logical_char.format.underline.color, 1))
                painter.drawPath(path)

            case _:
                raise ValueError(
                    f"Invalid underline style: {logical_char.format.underline.style!r}"
                )

    def _process_mode_info(self, mode_info):
        """Prepare a structure after mode_info for faster recurrent usage."""
        result_struct = {}
        mode_info = mode_info.copy()  # copy because will consume

        # this is just discarded (if present), as neovim documentation says "to be implemented"
        mode_info.pop("mouse_shape", None)

        # these two are also discarded as they are deprecated
        mode_info.pop("hl_id", None)
        mode_info.pop("id_lm", None)

        if "attr_id_lm" in mode_info:
            attr_id_lm = mode_info.pop("attr_id_lm")
            if attr_id_lm != 0:
                log_notdone("Cursor attribute id for when :lmap is on", attr_id_lm=attr_id_lm)

        if "blinkon" in mode_info:
            # consider that if one is present, the three will be
            blinkon = mode_info.pop("blinkon")
            blinkoff = mode_info.pop("blinkoff")
            blinkwait = mode_info.pop("blinkwait")
            if blinkon or blinkoff or blinkwait:
                log_notdone("Cursor blinking", on=blinkon, off=blinkoff, wait=blinkwait)

        # we need a cursor, be gentle with input and default to a full block
        cursor_shape = mode_info.pop("cursor_shape", "block")
        cursor_perc = mode_info.pop("cell_percentage", 20)
        cursor_attr_id = mode_info.pop("attr_id", 0)
        if cursor_shape == "block":
            cursor_painter = partial(self._paint_cursor_block, cursor_attr_id)
        elif cursor_shape == "vertical":
            cursor_painter = partial(self._paint_cursor_vertical, cursor_attr_id, cursor_perc)
        elif cursor_shape == "horizontal":
            cursor_painter = partial(self._paint_cursor_horizontal, cursor_attr_id, cursor_perc)
        else:
            log_notdone("Cursor with this shape", shape=cursor_shape, percentage=cursor_perc)
            return
        result_struct["cursor_painter"] = cursor_painter

        if mode_info:
            log_notdone("Some mode change info remained unprocessed", mode_info=mode_info)

        return result_struct

    def change_mode(self, mode_info):
        """Change mode to the specified information.

        The parsing for each mode info happens once, producing a cached structure to
        speed up recurrent calls with same info.
        """
        # get it from the cache or process
        cache_key = tuple(mode_info.items())
        struct = self.mode_info_structs.get(cache_key)
        if struct is None:
            struct = self._process_mode_info(mode_info)
            self.mode_info_structs[cache_key] = struct

        # use the structure
        self.cursor_painter = struct["cursor_painter"]

    def _build_text_format(self, hl_id: int | None) -> CharFormat:
        """Get the format for the text. If None, return default colors."""
        # check if in dynamic cache
        dyncache_labels = ("default_colors", "hl-attrs")
        fmt = self.main_window.nvim_notifs.dyncache.get(dyncache_labels, hl_id)
        if fmt is not None:
            return fmt

        # the base is always the default color
        default_colors = self.main_window.nvim_notifs.structs["default_colors"]
        fmt = CharFormat(
            background=QColor(default_colors["background"]),
            foreground=QColor(default_colors["foreground"]),
        )

        # 'special' is color for underline, this is the default, may be modified later
        special_color = QColor(default_colors["special"])

        if hl_id:  # cover also the case of it being 0, which *may* indicate default colors
            hl_attrs = self.main_window.nvim_notifs.structs["hl-attrs"]
            hl = hl_attrs[hl_id].copy()  # copy because will consume

            # basic set of attributes
            attr_names = ("foreground", "background", "strikethrough", "italic", "bold")
            for name in attr_names:
                if name in hl:
                    setattr(fmt, name, hl.pop(name))

            # colors may be reversed
            reverse = hl.pop("reverse", False)
            if reverse:
                fmt.foreground, fmt.background = fmt.background, fmt.foreground

            # XXX: we need to support 'url', but not sure the info that comes and how it spans

            # XXX: we need to support 'blend', but still not sure how
            #   blend: blend level (0-100). Could be used by UIs to support blending floating
            #   windows to the background or to signal a transparent cursor.

            # all variations of underlining; note the 'for' continues to the end (instead of
            # breaking on first find) because we want to "consume" all possible flags
            underline_style = None
            for style in UNDERLINE_STYLES:
                if hl.pop(style, False):
                    underline_style = style
            if underline_style is not None:
                if "special" in hl:
                    special_color = QColor(hl.pop("special"))
                fmt.underline = CharUnderline(color=special_color, style=underline_style)

            if hl:
                logger.warning("Some text format remained unprocessed: {}", hl)

        # set the format in dynamic cache
        self.main_window.nvim_notifs.dyncache.set(dyncache_labels, hl_id, fmt)
        return fmt

    def write_grid(self, row: int, col: int, sequence: list[Any]):
        """Write a sequence starting in the given row/column.

        The sequence is a list of text, or text and highlight id, or text, highlight id and
        repetitions.
        """
        textinfo = []

        fmt = self._build_text_format(None)
        for item in sequence:
            match item:
                case ['']:
                    # special case to indicate that the previous char is width
                    text = None
                case [text]:
                    hl_id = None
                case [text, hl_id]:
                    pass
                case [' ', 0, 0]:
                    # looks like used at the end of each sequence; looks not useful
                    continue
                case [text, hl_id, repeat]:
                    text = text * repeat
                case _:
                    raise ValueError(f"Wrong sequence format when writing to display: {item!r}")

            if hl_id is not None:
                # transform Neovim highlight into Qt format, caching it
                fmt = self.nvimhl_to_qtfmt.get(hl_id)
                if fmt is None:
                    fmt = self._build_text_format(hl_id)
                    self.nvimhl_to_qtfmt[hl_id] = fmt

            textinfo.append((text, fmt))

        self.lines.add(row, col, textinfo)
